#Main Code

#import python modules
import numpy as np
import time

#import compiled fortran library
import polylib

import atomwrite
import sys
import os

#whether or not to use the 3d visualization
UseVisual = False
if UseVisual:
    #import custom visualization library
    import atomvis

#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.


def CalcEnergy(Pos):
    """Wrapper for energy calculation.
Input:
    Pos: (N,3) position array
Output:
    PEnergy: potential energy, float
"""
    global L, kappa, lamba
    return polylib.calcenergy(Pos, L, kappa, lamba)


def CalcEnergyForces(Pos, Forces):
    """Wrapper for energy and force calculation.
Input:
    Pos: (N,3) position array
    Forces: (N,3) force array
Output:
    PEnergy: potential energy, float
    Forces: (N,3) force array (updated)
"""
    global L, kappa, lamba
    return polylib.calcenergyforces(Pos, L, kappa, lamba, Forces)


def LineSearch(Pos, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = Dir / np.sqrt(np.sum(Dir * Dir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [CalcEnergy(Pos + NormDir * x) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = CalcEnergy(Pos + NormDir * dx)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [CalcEnergy(Pos + NormDir * 2. * dx)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [CalcEnergy(Pos + NormDir * Dists[-1])]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = CalcEnergy(Pos + NormDir * d3)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = CalcEnergy(Pos + NormDir * d3)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]

    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosMin)
        else:
            #initialize the visualization window
            atomvis.Init(PosMin)
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    PE, Forces = CalcEnergyForces(Pos, np.zeros_like(Pos))
    Dir = Forces
    OldPE = 1.e300
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = CalcEnergyForces(Pos, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos


def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #make an array of lattice sites
    r = np.arange(L, dtype=float) - 0.5*L
    #loop through x, y positions in lattice until done
    #this will results in a snake like starting position
    #for every atom in the system
    i = 0
    for y in range(len(r)):
        if (y % 4 == 0):
           for x in r:
              Pos[i] = np.array([x,r[y],r[0]]) + 0.1 * (np.random.rand(3) - 0.5)
              Pos[i+N/2] = np.array([x,r[y],r[2]]) + 0.1 * (np.random.rand(3) - 0.5)
              i += 1
              if i >= N/2:
                  return Pos
        elif (y % 4 == 1):
           Pos[i] = np.array([r[len(r)-1],r[y],r[0]]) + 0.1 * (np.random.rand(3) - 0.5)
           Pos[i+N/2] = np.array([r[len(r)-1],r[y],r[2]]) + 0.1 * (np.random.rand(3) - 0.5)
           i += 1
           if i >= N/2:
               return Pos
        elif (y % 4 == 2):
           for x in r[::-1]:
              Pos[i] = np.array([x,r[y],r[0]]) + 0.1 * (np.random.rand(3) - 0.5)
              Pos[i+N/2] = np.array([x,r[y],r[2]]) + 0.1 * (np.random.rand(3) - 0.5)
              i += 1
              if i >= N/2:
                  return Pos
        else:
           Pos[i] = np.array([r[0],r[y],r[0]]) + 0.1 * (np.random.rand(3) - 0.5)
           Pos[i+N/2] = np.array([r[0],r[y],r[2]]) + 0.1 * (np.random.rand(3) - 0.5)
           i += 1
           if i >= N/2:
               return Pos
    print "Initial Positions failed"
    return Pos

def RescaleVelocities(Vel, T):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel)
    #find velocity scale factor from ratios of kinetic energy
    VScale = np.sqrt(1.5 * len(Vel) * T / KE)
    Vel = Vel * VScale
    return Vel  

def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T)
    return Vel


def VelocitiesDistribution(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    # Pick an atom
    MoveAtom = np.random.randint(len(Pos))
    # Choose its velocity from a Gaussian distribution
    Vel[MoveAtom,:] = np.random.normal(0,np.sqrt(T),size=3)
    return Vel

def InitAccel(Pos):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #get the acceleration from the forces
    PEnergy, Accel = CalcEnergyForces(Pos, np.zeros_like(Pos))
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))
 
def RadiusOfGyration(Pos):
    """Returns the radius of gyration squared for Pos"""
    Rg=0.0
    N=len(Pos)
    rbar = Pos.mean(axis=0)
    # find rbar, the mean position first
    for i in range(len(Pos)):
        Vec = Pos[i,:]
        Vec = Vec - rbar
        # Create a vector of r-rbar
        Rg += np.dot(Vec,Vec)
        # Add its square to the Rg variable
        Rg = Rg/N
    return Rg

#### GLOBAL VARIABLES ####
T = 1.2
dt = .001
rho = 2.2e-7
N = int(sys.argv[1])
lamba = float(sys.argv[2])
f = file(sys.argv[3], "w")
L = (N/(2*rho)) ** (1.0/3.0)
kappa = -1./(L/3.)*np.log(1.e-4*L/(3.*lamba))

f.write("Box Size: %8.3f & Kappa: %8.6f\n" % (L,kappa))

	
## Initialization of position
Pos = InitPositions(N, L)
PE, Pos = ConjugateGradient(Pos,0.001,1.e-8,1.e-10)
Vel = InitVelocities(N, T)
Accel = InitAccel(Pos)
# Two lines below used for movie making
#Names = ['N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N','O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O']
#Pdb = atomwrite.pdbfile("anim80_10.pdb",AtomNames = Names)

## Run

f.write("#KE\tPE\tTotal E\tTempinst\n")
RescaleFreq = 1500

## Equilibration
NStepsEquil =50050000
for i in range(0,NStepsEquil):
    
    if i % RescaleFreq == 0:
        Vel = VelocitiesDistribution(N, T)

    Pos, Vel, Accel, KE, PE = polylib.vvintegrate(Pos, Vel, Accel, L, kappa, lamba, dt)

    # Print out results
    if i % 10 == 0:
        f.write("%d \t %8.3f \t %8.3f \t %8.3f \t %8.3f \t %8.6f \n" % (i,KE,PE,KE+PE,InstTemp(Vel),RadiusOfGyration(Pos))) 

    # Lines below for movie making
    #if (i % 100 == 0 and i >= 100000 and i < 130000):
    #    Pdb.write(Pos - Pos.mean(axis=0))

    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(Pos)
        else:
            #initialize the visualization window
            atomvis.Init(Pos)

#Pdb.close()







